%############### PROVER MJCR
%

% #######################



:- module(proverMJCR, [
		buildProof/3, % buildProof(+Seq,-Proof,-ProofType : {proof, ref})
		search_main/2,
		is_proof/1,
	        is_ref/1
	      ]
	   ).

:- use_module(basic,[
		  is_propVar/1,
		  is_prime/1,
		  is_rposSubf/2,
	          get_rposSubf_list/2,
		  tr_formula_list/2
		  %pitpToProver/2,
		  ]).

:- use_module(syntax).



%############    SYNTAX  ############
%
%
%  *  Formula A,B ::=  P | #f |  A & B | A v B | A --> B
%
%     where:
%     - P is a Prolog atom and represents a PV (propositional variable)
%     -	#f represents false
%
%    Prime formulas: P | #f
%
%
%  *  unfocused_sequent	::=   Gamma ==> A @ Delta
%     focused-sequent ::=
%	 Gamma @ Theta @ foc(H) =*=> lab(Label) @ F @ Delta
%    (focus on H)
%
%    where:
%     -   A, H  are formulas
%     -   Gamma, Delta, Theta are lists of formulas
%     -   F is a prime formula
%     -   Label = 0,1
%
%  *  Proof ::=   proof(RootRule,RootSequent,PremiseList)
%
%     Let PremiseList = [Pr_1 ... Pr_n ] be a list of proofs.
%     Then, Proof represents the derivation
%
%
%	       Pr_1  .....  Pr_n
%	      ------------------- RootRule
%		RootSequent
%
%   The term
%
%        proof(leaf,Sequent,[])
%
%   represents a leaf of a tree
%
%   * Refutation ::=  ref(PropVarList)
%
%	  ref([P1,...., Pn]) represents an interpretation falsifying a sequent
%
%
%  Sf+(H) is the set of relevant positive subformulas of H
%  Sf+(Gamma) is the union of Sf+(H), for every H in Gamma



% is_irrSeq(+ Gamma  ==> A @ Delta)
%
% Check if Gamma  ==> A @ Delta  is an irreducible sequent

is_irrSeq(Gamma  ==> A @ Delta) :-
	get_rposSubf_list(Gamma,PosSfGamma),
        % PosSfGamma = Sf+(Gamma) = { Z | Z is a pos. subfurmula of a formula in Gamma }
        %% COND 1: every formula in Delta U {A}  is prime
	union(Delta,[A],Delta1),
	forall(member(D,Delta1), is_prime(D)),
	%% COND 2: ( Delta U {A,#f} ) /\  Sf+(Gamma) = \emptyset
	union(Delta1,[#f],Delta2),
        forall(member(D,Delta2), not(member(D,PosSfGamma))),
	%% COND 3: no D0 v D1 in Sf+(Gamma)
	not(member(_D0 v _D1, PosSfGamma)).

%%%  is_proof(+T)
%
%   Check if T is a proof

is_proof(proof(_Rule,_Seq,_PremiseList)).

%%%  is_ref(+T)
%
%   Check if T is a refutation
%

is_ref(ref(_)).

%##############    MAIN    ####################à

% buildProof( +Gamma ==> A @ Delta, -Proof, -ProofType)
%
%  Proof: a proof or a refutation of  Gamma ==> A @ Delta
%  ProofType in {proof, ref}


buildProof( Gamma ==> A @  Delta , Proof, ProofType) :- !,
	tr_formula_list([A| Gamma],[A1| Gamma1]),
	tr_formula_list(Delta,  Delta1),
	go_search_unfoc(Gamma1==> A1 @ Delta1,  Proof),
	(
	    is_proof(Proof) ->
	       ProofType = proof
	        ;
	      assertion(is_ref(Proof)),
	      ProofType = ref
	).






%#########################################################
%#############	 PROOF-SEARCH ALGORITHM     ##############
%#########################################################

%%   **NOTE** :
%%   henceforth we assume that formulas do not contain
%%   derived operators ~, <-->


%########   GOAL ORIENTED SEARCH #########

% search_main( +Gamma ==> A, -Result)
%
%  Result: a derivation or a refutation of Gamma ==> A



search_main( Gamma ==> A  , Result) :-
	go_search_unfoc( Gamma  ==> A @ [] ,  Result).




% go_search_unfoc( +Gamma ==> A @ Delta, -Result)
%
%  Result: a derivation or a refutation of Gamma ==> A @ Delta

% F prime - apply lf
go_search_unfoc(  Gamma  ==> F @ Delta, Result) :-
	is_prime(F) , !,
	RootSeq = ( Gamma  ==> F @ Delta ),
        ( %% apply lf1
          %  #f in Sf+(Gamma)
	  ( member(H,Gamma), is_rposSubf(#f,H))  ->
	      subtract(Gamma,[H],Gamma_H),  %% Gamma_H = Gamma \ {H}
	      go_search_foc(1, #f,Gamma_H @ [] @ foc(H) =*=> lab(1) @ F @ Delta, Result0),!,
	     (
	       is_proof(Result0) ->
		Result = proof(lf1,RootSeq, [Result0])
		 ;
		assertion(is_ref(Result0)),
		Result =  Result0
	      )
	;
	  %% apply lf0
	   % F is a PV and F in Sf+(Gamma)
	  ( is_propVar(F), member(H,Gamma), is_rposSubf(F,H) )  ->
	      subtract(Gamma,[H],Gamma_H),  %% Gamma_H = Gamma \ {H}
	      go_search_foc(0,F, Gamma_H @ [] @ foc(H) =*=> lab(0) @ F @ Delta, Result0),!,
	      (
	      is_proof(Result0) ->
		  Result = proof(lf0,RootSeq, [Result0])
		  ;
		  assertion(is_ref(Result0)),
                  Result =  Result0
	      )
	 ;
	  % there exists A v B in Sf+(Gamma)
	     ( member(H,Gamma), is_rposSubf(A v B,H) ) ->
	       subtract(Gamma,[H],Gamma_H),  %% Gamma_H = Gamma \ {H}
	      go_search_foc(0,A v B, Gamma_H @ [] @ foc(H) =*=> lab(0) @ F @ Delta, Result0),!,
	      (
	      is_proof(Result0) ->
		  Result = proof(lf0,RootSeq, [Result0])
		  ;
                assertion(is_ref(Result0)),
		Result =  Result0
	      )
	  ;
	  %% apply rlf (rest+lf)
          % there exists P s.t. P is a PV, P is in Delta and P is in Sf+(Gamma)
	  ( member(P,Delta), is_propVar(P), member(H,Gamma), is_rposSubf(P,H) ) ->
	    subtract(Gamma,[H],Gamma_H), %% Gamma_H = Gamma \ {H}
	    subtract(Delta,[P],Delta_P), %% Delta_P = Delta \ {P}
	    union([F],Delta_P,FDelta_P), %%  FDelta_P =  (Delta \ {P})  U {F}
	    go_search_foc(0,P, Gamma_H @ [] @ foc(H) =*=> lab(0) @ P @ FDelta_P, Result0),!,
	    (
	      is_proof(Result0) ->
		  Result = proof(rlf,RootSeq, [Result0])
		  ;
		assertion(is_ref(Result0)),
		Result =  Result0
	      )
	  ;
	  %% apply rest
          % restart with a non-prime formula D in Delta
	  ( member(D,Delta), not(is_prime(D)) ) ->
	    subtract(Delta,[D],Delta_D),  %% Delta_D = Delta \ {D}
	    union([F],Delta_D,FDelta_D),  %% FDelta_D =  (Delta \ {D})  U {F}
	    go_search_unfoc(Gamma  ==> D @ FDelta_D, Result0),!,
	    (
	       is_proof(Result0) ->
		Result = proof(rest,RootSeq, [Result0])
		 ;
		assertion(is_ref(Result0)),
		Result =  Result0
	      )

         ; % if none of the above conditions is true, the input sequent is irreducible
	   % A countermodel is the set of PV in Sf+(Gamma)
	  assertion(is_irrSeq(RootSeq) ),
	  get_rposSubf_list(Gamma,PosSubfGamma),
	  findall(P, (member(P,PosSubfGamma),is_propVar(P)),PropVarList),
	 %% NOTE: setof fails if PosSubfGamma is empty!
	  Result = ref(PropVarList)
	).


% apply rAnd
go_search_unfoc(Gamma  ==> A & B @ Delta, Result) :-
        go_search_unfoc(Gamma  ==> A @ Delta, ResultA), !,
	(
	   is_ref(ResultA) ->
	      Result = ResultA
	;
	    assertion(is_proof(ResultA)),
	    go_search_unfoc(Gamma  ==> B @ Delta, ResultB), !,
	    (
		 is_ref(ResultB) ->
		    Result = ResultB
		    ;
		    assertion(is_proof(ResultB)),
		    Result = proof(rAnd , Gamma  ==> A & B @ Delta,  [ResultA,ResultB])
	    )
	).


% apply rOr
go_search_unfoc(  Gamma  ==> A v B @ Delta, Result) :-
        union(Delta,[B], BDelta),
	go_search_unfoc(Gamma  ==> A @ BDelta, Result0), !,
	(
	   is_proof(Result0) ->
	      Result = proof(rOr , Gamma  ==> A v B @ Delta,  [Result0])
	      ;
	      assertion(is_ref(Result0)),
	      Result =  Result0
	).

% apply rImpl
go_search_unfoc(  Gamma  ==> A-->B @ Delta, Result) :-
	union(Gamma,[A],AGamma),
        go_search_unfoc(AGamma  ==> B @ Delta, Result0), !,
	(
	   is_proof(Result0) ->
	      Result = proof(rImpl, Gamma  ==> A-->B @ Delta, [Result0] )
	;
	   assertion(is_ref(Result0)),
	   Result =  Result0
	).

% go_search_foc(+Label, +Goal, +FocusedSequent, -Result).
%
% Resulat: a proof of the focused sequent FocusedSequent or null
% Label: the label of the sequent (0,1)
% Goal: the focused formula to be obtained at the end of focus


% go_search_foc(Label,Target, Gamma @ Theta @ foc(H) =*=> lab(Label) @ F @Delta, Result).

%%  apply ax
go_search_foc(0,F, Gamma @ Theta @ foc(F) =*=> lab(0) @ F @ Delta, Result):-
	is_propVar(F), !,
	Result = proof(ax,  Gamma @ Theta @ foc(F) =*=> lab(0) @ F@ Delta, []).


%  apply lBot
go_search_foc(1,#f, Gamma @ Theta @ foc(#f) =*=> lab(1) @ F @ Delta, Result):- !,
	Result = proof(lBot,  Gamma @ Theta @ foc(#f) =*=> lab(1) @ F@ Delta, []).

%% apply lAnd

go_search_foc(Label,T, Gamma @ Theta @ foc(A & B) =*=> lab(Label) @ F @ Delta, Result):- !,
   RootSeq =  Gamma @ Theta @ foc(A & B) =*=> lab(Label) @ F @ Delta,
 (
    is_rposSubf(T,A) ->
       union(Theta,[B],BTheta),
       go_search_foc(Label,T, Gamma @ BTheta @ foc(A) =*=> lab(Label) @ F @ Delta, ResultA), !,
	(
	   is_proof(ResultA) ->
	    Result = proof(lAnd0,RootSeq,[ResultA])
	   ;
	   assertion(is_ref(ResultA)),
	   Result =  ResultA
       )
    ;
    assertion(is_rposSubf(T,B)),
    union(Theta,[A],ATheta),
    go_search_foc(Label,T, Gamma @ ATheta @ foc(B) =*=> lab(Label) @ F @ Delta, ResultB), !,
	(
	   is_proof(ResultB) ->
	    Result = proof(lAnd1,RootSeq,[ResultB])
	   ;
	   assertion(is_ref(ResultB)),
	   Result =  ResultB
	)
).

%% apply lOr

go_search_foc(0,A v B, Gamma @ Theta @ foc(A v B) =*=> lab(0) @ F @ Delta, Result):- !,
	union(Gamma,Theta,GammaUTheta),
	  union(GammaUTheta, [A],  AGammaUTheta),
	  go_search_unfoc(AGammaUTheta ==> F @ Delta,ResultA), !,
	  (
	    is_ref(ResultA) ->
	        Result = ResultA
	    ;
            assertion(is_proof(ResultA)),
	    union(GammaUTheta, [B],  BGammaUTheta),
	    go_search_unfoc(BGammaUTheta ==> F @ Delta,ResultB), !,
	    (
	    is_ref(ResultB) ->
	      Result = ResultB
	    ;
	       assertion(is_proof(ResultB)),
               Result = proof(lOr, Gamma @ Theta @ foc(A v B) =*=> lab(0) @ F @ Delta, [ResultA,ResultB])
            )
	 ).


%% apply lImpl

go_search_foc(Label,T, Gamma @ Theta @ foc(A --> B) =*=> lab(Label) @ F @ Delta, Result):- !,
	union(Gamma,Theta,GammaUTheta),
	union(Delta,[F],FDelta),
	  go_search_unfoc(GammaUTheta ==> A @ FDelta,ResultA), !,
	  (
	    is_ref(ResultA) ->
	        Result = ResultA
	    ;
	    assertion(is_proof(ResultA)),
            go_search_foc(Label,T, Gamma @ Theta @ foc(B) =*=> lab(Label) @ F @ Delta, ResultB), !,
	    (
	    is_ref(ResultB) ->
	      Result = ResultB
	    ;
               assertion(is_proof(ResultB)),
               Result = proof(lImpl, Gamma @ Theta @ foc(A--> B) =*=> lab(Label) @ F @ Delta, [ResultA,ResultB])
            )
	 ).











































