%################### TRANSLATION THETA : MJCR ---> NC #################

:- module(translationTheta,
	  [
	      theta/2
	  ]).

:- use_module(syntax).


%     SYNTAX
%
%   *  A marked formula is represented by the operator ^, e.g.
%      "D marked" is represented by the Prolog term
%
%            D^
%
%   *  A pattern
%
%             Gamma ; H '|>'
%
%    is represented by the Prolog term
%
%           Gamma @ H '|>'
%
%
%  * A ptree is an MJC-tree of the form
%
%     { Gamma @ foc(A) '|>' }
%		.
%		.
%	        .
%       Gamma @ foc(H)  '|>'
%
%   where:
%    -   the pattern Gamma @ foc(H)  '|>' is the root
%    -   the pattern Gamma @ foc(A) '|>'  is the only open leaf.
%    -   the displayed branch only contains patterns
%
% * NC-sequents are represented by the Prolog terms
%
%   Gamma '|--' up(A)            Gamma '|--' down(A)
%
%
% * A down-open tree is an NC-tree of the form
%
%
%    { Gamma  '|--' down(H) }
%		.
%		.
%	        .
%       Gamma  '|--' down(A)
%
%   where:
%   - the root is Gamma  '|--' down(A)
%   - the only open leaf is  Gamma  '|--' down(H)
%   - the displayed branch only contais down-sequents
%
%

%===========   AUXILIARY PREDICATES =================




% split(+MJC_Proof, -MJC_Proof_out ,-MJC_Open_Tree_out)
%
% MJC_Proof: an MJC_Proof having as root rule R one of
%    lc0, lc1,lAnd0,lAnd1, lImpl,lor,ax,lBot
%
% MJC_Proof is split in MJC_Proof_out and MJC_Open_Tree_out such that:
%
% o MJC_Open_Tre_out has root rule R and one only one open-leaf
% o MJC_Open_Tree_out only contains applications of
%    lc0, lc1, lAnd0, lAnd1, lImpl
%  o MJC_Proof can be obtained by replacing the open leaf of
%      MJC_Open_Tree_out with MJC_Proof_out
%
%                                               { sigma_0}
%                     MJC_Proof_out        MJC_Open_Tree_out
%                         sigma_0                sigma_1
%     MJC_Proof  = - - - - - - - - - - - - - - - - - - - - - - -
%       sigma		     sigma
%


% ax,lOr,lBot
split(proof(Rule,AxiomSequent,[]), proof(Rule,AxiomSequent,[]), proof(leaf, AxiomSequent,[])) :- !,
    assertion(member(Rule,[ax,lOr,lBot])).

% lc0, lc1,lAnd0,lAnd1
split(proof(lOr, RootSeq, Prems), proof(lOr,RootSeq,Prems), proof(leaf, RootSeq,[])) :- !.

split(proof(Rule,RootSequent,[P0]),  Proof_out, DownOpenTree) :- !,
	assertion(member(Rule,[lc0, lc1,lAnd0,lAnd1])),
	split(P0, Proof_out, DownOpenTree0),
	DownOpenTree = proof(Rule,RootSequent,[DownOpenTree0]).

%  lImpl
split(proof(Rule,RootSequent,[P0,P1]),  Proof_out, DownOpenTree) :-
	assertion(Rule = lImpl),
	split(P1, Proof_out, DownOpenTree1),
	DownOpenTree = proof(Rule,RootSequent,[P0,DownOpenTree1]).

% translateRule(+MJC_Rule, -NC_Rule).

translateRule(rAnd,andI).
translateRule(rOr0,or0I).
translateRule(rOr1,or1I).
translateRule(rImpl,implI).
translateRule(lNeg,negE).
translateRule(rc,markE).
translateRule(rOrp,orpI).

% abstract_seq(+Focused_Seq, -Pattern)
%  Focused_Seq can be obtained by filling the holes of Pattern
%

abstract_seq(Gamma @foc(A) =*=> lab(_L) @ _F, Gamma @foc(A) '|>').



% abstract_tree(+Focused_Open_MJC_tree, -MJC_Ptree_out)
%
%  MJC_Ptree is the ptree obtained by replacing
%  every focused-sequent Sigma in the main-branch
%  with the corresponding pattern

abstract_tree(proof(leaf, Seq, []),  proof(leaf, Pattern, [])) :-
	abstract_seq(Seq,Pattern).

abstract_tree(proof(ax, Seq, []), proof(ax, Pattern, []) ) :-
	abstract_seq(Seq,Pattern).

% lAnd0,lAnd1
abstract_tree(proof(LAnd, Seq, [T0]), PTree_out) :- !,
	assertion(member(LAnd,[lAnd0,lAnd1])),
	abstract_tree(T0,PT0),
	abstract_seq(Seq,Pattern),
	PTree_out = proof(LAnd, Pattern, [PT0]).

% lImpl
abstract_tree(proof(LImpl, Seq, [P0,T1]), PTree_out) :- !,
	assertion(LImpl = lImpl),
	abstract_tree(T1,PT1),
	abstract_seq(Seq,Pattern),
	PTree_out = proof(LImpl, Pattern, [P0,PT1]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%  The	 maps theta    and thetaPrime
%

% theta(+MJC_Proof,-NC_Proof)

theta(proof(LCRule, Gamma ==> K, [P]), Proof_out) :-
	member(LCRule,[lc0,lc1]),!,
	split( proof(LCRule, Gamma ==> K, [P]) ,P0, LC_Tree),
	LC_Tree =   proof(LCRule,_,[T0]),
	assertion(member(LCRule,[lc0,lc1])),
	T0 = proof(_, _ @ foc(H) =*=> _,_Prems),  %% extract H
	abstract_tree(T0,Abst_T0),
	thetaPrime(Abst_T0,ThetaPrime_Abst_T0),
	RootSeq =  Gamma '|--' up(K),
	% by cases on the root rule of P0
	(
	    P0 = proof(ax,_,_) ->
	    ProofCoerc =  proof(coerc, RootSeq,[ThetaPrime_Abst_T0]),
	    ProofID = proof(id, Gamma '|--' down(H),[]),
	    replace_leaf(ProofID,  ProofCoerc,Proof_out)
	 ;
	    P0 = proof(lOr, _ ,  [P1,P2]) ->
	    theta(P1,Theta_P1),
	    theta(P2,Theta_P2),
	    ProofID = proof(id, Gamma '|--' down(H),[]),
	    ProofOrE = proof(orE, RootSeq, [ ThetaPrime_Abst_T0,Theta_P1,Theta_P2]),
	    replace_leaf(ProofID,ProofOrE,Proof_out)
	;
	   P0 = proof(lBot,_,_) ->
		union(Gamma, [K^], KMarked_U_Gamma),
		ProofID = proof(id,  KMarked_U_Gamma  '|--' down(H),[]),
	        ProofBoteCl = proof(boteCl, RootSeq,[ThetaPrime_Abst_T0]),
		replace_leaf(ProofID,ProofBoteCl,Proof_out)
	).



theta(proof(Rule, Gamma ==> K, Prems), Proof_out) :-
	assertion(member(Rule,[rAnd,rOr0,rOr1,rImpl,rc,rOrp])),
	theta_list(Prems,Prems1),
	translateRule(Rule,Rule1),
	Proof_out = proof(Rule1, Gamma '|--' up(K), Prems1).


theta_list([],[]).
theta_list([Proof|Proofs],[Proof1|Proofs1]) :-
	   theta(Proof,Proof1),
	   theta_list(Proofs,Proofs1).



% thetaPrime(+MJC_PTree,-Down_Open_NC_Tree)

thetaPrime( proof(leaf, Gamma @ foc(H) '|>',[]),   proof(leaf, Gamma '|--'  down(H) ,[])   ) :-  !.


thetaPrime(proof(LAnd,  Gamma @ foc(H0 & H1) '|>',[PT]), NC_Tree_out) :- !,
	 assertion(member(LAnd,[lAnd0,lAnd1])),
	 (  LAnd = lAnd0 ->
	       H_k = H0,
	       AndE = and0E
	 ;
	      H_k = H1,
	      AndE = and1E
	 ),
	 OpenLeaf = proof(leaf,   Gamma '|--'  down(H0 & H1),[]),
	 TreeAndE = proof(AndE,  Gamma '|--'  down(H_k) , [OpenLeaf]),
	 thetaPrime(PT,ThetaPrime_PT),
	 replace_leaf(TreeAndE,ThetaPrime_PT,NC_Tree_out).


thetaPrime(proof(lImpl,  Gamma @ foc(H0 --> H1) '|>',[P0,PT1]), NC_Tree_out) :- !,
	OpenLeaf = proof(leaf,   Gamma '|--'  down(H0 --> H1),[]),
	 theta(P0,Theta_P0),
	 TreeImplE = proof(implE,  Gamma '|--'  down(H1) , [OpenLeaf,Theta_P0]),
	 thetaPrime(PT1,ThetaPrime_PT1),
	 replace_leaf(TreeImplE,ThetaPrime_PT1,NC_Tree_out).




replace_leaf(P0, proof(leaf,_,[]), P0):- !.

replace_leaf(P0, proof(Rule,RootSeq,Prems),Proof_out) :-
        replace_leaf_list(P0,Prems,Prems_r),
        Proof_out =  proof(Rule,RootSeq,Prems_r).

replace_leaf_list(_P0,[],[]) :- !.

replace_leaf_list(P0, [Proof |Proofs] , [Proof_r|Proofs_r] ) :-
           replace_leaf(P0,Proof,Proof_r),
           replace_leaf_list(P0, Proofs , Proofs_r ).


























